'use strict';

Object.defineProperty(exports, "__esModule", {
    value: true
});
exports.strategy = undefined;

var _core = require('../../../../../core.js');

var _core2 = _interopRequireDefault(_core);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var strategy = {
    name: 'single',
    condition: function condition(provider, convertedItems) {
        var requests = convertedItems.getItems();
        return requests.length > 0;
    },
    save: function save(provider, convertedItems, callBack) {
        var that = provider;
        var items = convertedItems.getItems();

        var doSave = function doSave(items, index, done) {
            var item = items[index];
            if (!item) return done();

            var request = item.request.build().get();
            var requestData = [request, function (data, response) {
                if (response.statusCode >= 200 && response.statusCode < 300) {
                    var item = convertedItems.getByResponse(response, index);
                    if (item instanceof _core2.default.Entity && response.statusCode != 204) {
                        that.reload_fromResponse(item, data, response);
                        convertedItems.setProcessed(item);
                    }

                    doSave(items, ++index, done);
                } else {
                    done(response);
                }
            }, done];

            that.appendBasicAuth(requestData[0], that.providerConfiguration.user, that.providerConfiguration.password, that.providerConfiguration.withCredentials);
            that.context.prepareRequest.call(that, requestData);
            that.oData.request.apply(that, requestData);
        };

        doSave(items, 0, function (err, result) {
            if (err) return callBack.error(that.parseError(err));
            callBack.success(result);
        });
    }
};

exports.strategy = strategy;